@@ -12,7 +12,7 @@
 # GNU Affero General Public License for more details.
 #
 # You should have received a copy of the GNU Affero General Public
-# License, version 3, along with BANG.  If not, see
+# License, version 3, along with yldprolog.  If not, see
 # <http://www.gnu.org/licenses/>
 #
 # Copyright 2018-2019 - Tim Hemel
@@ -157,6 +157,8 @@ class Answer:
         self.values = values
     def match(self, args):
         return unify_arrays(args, self.values)
+    def __str__(self):
+        return f'Answer({[to_python(x) for x in self.values]})'
 
 class YPFail(object):
     """Iterator that always stops."""
@@ -307,21 +309,91 @@ class YP(object):
         for y in unify(bag, results):
             yield False
 
-    def call(self, goal):
-        '''call/1 calls goal.'''
-        yield from self.query(goal._name, goal._args)
+    def call(self,goal,*args):
+        '''call/n (:Goal, Arg, ...). Calls Goal with args appended to its arguments.'''
+        goal_value = get_value(goal)
+        if isinstance(goal_value, Atom):
+            goal_name = to_python(goal_value)
+            goal_args = []
+        elif isinstance(goal_value, Functor):
+            goal_name = goal._name
+            goal_args = goal._args
+        else:
+            # TODO: raise exception
+            pass
+        yield from self.query(goal_name, goal_args + list(args))
 
     def once(self, goal):
         '''once/1 calls goal only once.'''
         q = self.call(goal)
         yield next(q)
 
+    def asserta(self, term):
+        '''asserta(Term) adds Term to the facts database at the beginning.'''
+        if isinstance(term, Functor):
+            self.assert_fact(self.atom(term._name), term._args, False)
+        elif isinstance(term, Atom):
+            self.assert_fact(term, [], False)
+        return YPSuccess()
+
+    def assertz(self, term):
+        '''assertz(Term) adds Term to the facts database at the end.'''
+        if isinstance(term, Functor):
+            self.assert_fact(self.atom(term._name), term._args)
+        elif isinstance(term, Atom):
+            self.assert_fact(term, [])
+        return YPSuccess()
+
+    def retract(self, term):
+        '''retract(Term) removes all dynamic facts matching Term and backtracks over identical clauses.'''
+        if isinstance(term, Functor):
+            name = term._name
+            args = term._args
+        elif isinstance(term, Atom):
+            name = term
+            args = []
+
+        remaining_clauses = self._find_predicates(name, len(args))[:]
+        i = 0
+        while i < len(remaining_clauses):
+            clause = remaining_clauses[i]
+            match = False
+            for cut in clause.match(args):
+                match = True
+                del remaining_clauses[i]
+                self._update_predicate(self.atom(name), len(args), remaining_clauses)
+                yield False
+            if not match:
+                i += 1
+
+    def retractall(self, term):
+        '''retractall(Term) removes all dynamic facts matching Term, without backtracking over identical clauses.'''
+        if isinstance(term, Functor):
+            name = term._name
+            args = term._args
+        elif isinstance(term, Atom):
+            name = term
+            args = []
+        remaining_clauses = []
+        for clause in self._find_predicates(name, len(args)):
+            match = False
+            for cut in clause.match(args):
+                    match = True
+            if not match:
+                remaining_clauses.append(clause)
+        self._update_predicate(self.atom(name), len(args), remaining_clauses)
+        return YPSuccess()
+
     def _set_builtin_predicates(self):
         self.register_function('=', builtin_eq)
         self.register_function('\\=', self.builtin_neq)
         self.register_function('findall', self.findall)
-        self.register_function('call', self.call)
+        self.register_function('call', self.call, arity=-1)
         self.register_function('once', self.once)
+        self.register_function('assertz', self.assertz)
+        self.register_function('asserta', self.asserta)
+        self.register_function('retract', self.retract)
+        self.register_function('retractall', self.retractall)
 
     def clear(self):
         """clears all defined atoms, variables, facts and rules."""
@@ -336,9 +408,11 @@ class YP(object):
         """
         self._atom_store.setdefault(name, Atom(name))
         return self._atom_store[name]
+
     def variable(self):
         """Create a variable in this engine."""
         return Variable()
+
     def functor(self, name, args):
         """Create a functor in this engine with name name and the list of Prolog terms args
         as the functor arguments."""
@@ -352,6 +426,7 @@ class YP(object):
     def functor3(self, name, arg1, arg2, arg3):
         """Compatibility function for creating a functor with three arguments."""
         return Functor(name, [arg1, arg2, arg3])
+
     def listpair(self, head, tail):
         """Creates a Prolog listpair representing [head|tail]. An empty list is
         represented by the ATOM_NIL object member.
@@ -360,6 +435,7 @@ class YP(object):
             prologlist = yp.listpair(yp.atom('a'),yp.listpair(yp.atom('b'),yp.ATOM_NIL))
         """
         return Functor(self.ATOM_DOT, [head, tail])
+
     def makelist(self, l):
         """Creates a Prolog list from a Python list. l is a list of Prolog terms."""
         r = functools.reduce(lambda x, y: self.listpair(y, x), reversed(l), self.ATOM_NIL)
@@ -394,15 +470,22 @@ class YP(object):
         with open(fn, "r") as f:
             self.load_script_from_string(f.read(), fn=fn, overwrite=overwrite)
     
-    def register_function(self, name, func):
+    def register_function(self, name, func, arity=None):
         """Makes the function func available to the engine with Prolog name name. This can
         be used to call custom functions, implemented in Python. These function will have
         to behave as Prolog functions, i.e. they will need to yield boolean values.
         Since yldprolog 1.1.0, the arity is appended to name, allowing you to register
         functions with the same, but different arities.
+        If arity is None, the arity will be determined from the number of function arguments.
+        If arity is a non-negative integer, that arity will be used.
+        If arity is a negative integer, the function has a variable arity.
         """
-        arity = len(inspect.signature(func).parameters)
-        self.eval_context[f'{name}_{arity}'] = func
+        if arity is None:
+            arity = len(inspect.signature(func).parameters)
+        if arity < 0:
+            self.eval_context[f'{name}_n'] = func
+        else:
+            self.eval_context[f'{name}_{arity}'] = func
 
     def _find_predicates(self, name, arity):
         try:
@@ -413,21 +496,21 @@ class YP(object):
     def _update_predicate(self, name, arity, clauses):
         self._predicates_store[(name.name(), arity)] = clauses
 
-    def assert_fact(self, name, values):
-        """From the original YieldProlog:
-
-        assert values at the end of the set of facts for the predicate with the
-        name "name" and the arity len(values).
-        "name" must be an Atom.
-        values cannot contain unbound variables.
-        """
+    def assert_fact(self, name, values, append=True):
+        '''insert name(values) in the set of facts. If append is False, insert the
+        fact at the beginning, otherwise at the end.'''
         try:
             clauses = self._find_predicates(name.name(), len(values))
             # indexedanswers
         except YPException as e:
             clauses = []
-        clauses.append(Answer(values))
+        answer = Answer([get_value(v) for v in values])
+        if append:
+            clauses.append(answer)
+        else:
+            clauses.insert(0, answer)
         self._update_predicate(name, len(values), clauses)
+
     def query(self, name, args):
         """Creates a Prolog query for the symbol name, with arguments args. name is a string,
         args is a list of Prolog terms. The query will only be constructed, but not evaluated.
@@ -442,16 +525,11 @@ class YP(object):
         r = [ V1.get_value() for r in q ]
         assert r == [ yp.atom('tom') ]
         """
-        try:
-            if name not in self.eval_blacklist:
-                function = self.eval_context[f'{name}_{len(args)}']
-                return function(*args)
-        except KeyError as e:
-            pass
-        except TypeError as e: # args not matching
-            pass
-
-        return self.match_dynamic(self.atom(name), args)
+        yield from self.match_dynamic(self.atom(name), args)
+        if name not in self.eval_blacklist:
+            function = self.eval_context.get(f'{name}_{len(args)}', self.eval_context.get(f'{name}_n'))
+            if function is not None:
+                yield from function(*args)
 
     def evaluate_bounded(self, query, projection_function, recursion_limit=200):
         """Evaluates a query, but limits the recursion depth to recursion_limit. If a query
@@ -493,7 +571,7 @@ class YP(object):
         try:
             clauses = self._find_predicates(name.name(), len(args))
             return self._match_all_clauses(clauses, args)
-        except YPException:
+        except YPException as e:
             return YPFail()
 
     def _match_all_clauses(self, clauses, args):
