@@ -12,7 +12,7 @@
 # GNU Affero General Public License for more details.
 #
 # You should have received a copy of the GNU Affero General Public
-# License, version 3, along with BANG.  If not, see
+# License, version 3, along with yldprolog.  If not, see
 # <http://www.gnu.org/licenses/>
 #
 # Copyright 2018-2019 - Tim Hemel
@@ -306,6 +306,36 @@ def test_user_defined_function():
     r = list(q)
     assert side_effects == ['not blue']
 
+def test_register_function_variable_arities():
+    yp = YP()
+    side_effects = []
+    def func(arg1, *args):
+        side_effects.extend(to_python(x) for x in args)
+        for l1 in unify(arg1, yp.atom("blue")):
+            yield False
+    yp.register_function('func', func, arity=-1)
+    q = yp.query('func', [yp.atom('blue'), yp.atom('yellow'), yp.atom('green')])
+    r = list(q)
+    assert side_effects == ['yellow', 'green']
+
+def test_facts_and_function():
+    s = compile_prolog_from_string('''
+    person(mike).
+    person(joe).
+    person(pete).
+    person(zack).
+    ''', TestContext)
+
+    yp = YP()
+    yp.assert_fact(yp.atom('person'), [yp.atom('tom')])
+    yp.assert_fact(yp.atom('person'), [yp.atom('jerry')])
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('person', [v])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'tom', 'jerry', 'mike', 'joe', 'pete', 'zack' ]
+
+
 def test_run_infinite_script(get_compiled_file):
     yp = YP()
     yp.load_script_from_file(get_compiled_file('monkey.prolog'))
@@ -463,6 +493,14 @@ def test_builtin_predicate_call():
     result3 = [ to_python(v_child) for x in q ]
     assert result3 == result1
 
+    q = yp.query('call', [yp.functor('age',[v_child]), 5])
+    result4 = [ to_python(v_child) for x in q ]
+    assert result4 == result1
+
+    q = yp.query('call', [yp.atom('age'), v_child, 5])
+    result5 = [ to_python(v_child) for x in q ]
+    assert result5 == result1
+
 def test_builtin_predicate_once():
     s = compile_prolog_from_string('''
     age(peter, 7).
@@ -484,6 +522,116 @@ def test_builtin_predicate_once():
     result2 = [ to_python(v_child) for x in q ]
     assert result2 == result1
 
+def test_builtin_predicate_assertaz():
+    s = compile_prolog_from_string('''
+    initp() :- assertz(p(b)), assertz(p(a)), asserta(p(c)).
+    isp(X) :- initp() , p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('isp', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'c', 'b', 'a' ]
+
+def test_builtin_predicate_asserta_multiple():
+    s = compile_prolog_from_string('''
+    initp() :- assertz(p(a)), assertz(p(a)), assertz(p(b)), asserta(p(c)).
+    isp(X) :- initp() , p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('isp', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'c', 'a', 'a', 'b' ]
+
+def test_builtin_predicate_assertz_instantiated_variable():
+    s = compile_prolog_from_string('''
+    name(tom).
+    name(jerry).
+    initp() :- name(Y), assertz(p(Y)).
+    isp(X) :- p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    q = yp.query('initp', [ ])
+    r = list(q)
+    v = yp.variable()
+    q = yp.query('isp', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'tom', 'jerry' ]
+
+def test_builtin_predicate_assertz_free_variable():
+    s = compile_prolog_from_string('''
+    name(tom).
+    name(jerry).
+    initp() :- assertz(p(_)).
+    isp(X) :- name(X), p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    q = yp.query('initp', [ ])
+    r = list(q)
+    v = yp.variable()
+    q = yp.query('isp', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'tom', 'jerry' ]
+
+def test_builtin_predicate_retract_single_atom():
+    s = compile_prolog_from_string('''
+    foo(X) :- assertz(p(b)), assertz(p(a)), asserta(p(c)), retract(p(a)), p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('foo', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'c', 'b' ]
+
+def test_builtin_predicate_retract_duplicate_atom():
+    s = compile_prolog_from_string('''
+    foo(X) :- assertz(p(b)), assertz(p(a)), assertz(p(a)), asserta(p(c)), retract(p(a)), p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('foo', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'c', 'b', 'a', 'c', 'b' ]
+
+def test_builtin_predicate_retract_var():
+    s = compile_prolog_from_string('''
+    foo(X) :- assertz(p(b)), assertz(p(a)), assertz(p(a)), asserta(p(c)), retract(p(_)), p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('foo', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'b', 'a', 'a', 'a', 'a', 'a' ]
+
+def test_builtin_predicate_retractall_atom():
+    s = compile_prolog_from_string('''
+    foo(X) :- assertz(p(b)), assertz(p(a)), assertz(p(a)), asserta(p(c)), retractall(p(a)), p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('foo', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ 'c', 'b' ]
+
+def test_builtin_predicate_retractall_var():
+    s = compile_prolog_from_string('''
+    foo(X) :- assertz(p(b)), assertz(p(a)), assertz(p(a)), asserta(p(c)), retractall(p(_)), p(X).
+    ''', TestContext)
+    yp = YP()
+    yp.load_script_from_string(s, overwrite=False)
+    v = yp.variable()
+    q = yp.query('foo', [ v ])
+    r = [ to_python(v) for x in q ]
+    assert r == [ ]
 
 def test_load_scripts_with_dependencies_in_order(get_compiled_file):
     yp = YP()
